@@ -1,15 +1,19 @@
-from __future__ import with_statement
-
+from __future__ import annotations
 import datetime
 import json
 import logging
+import os
 import threading
+import inspect
+import types
+import typing
 
 from random import random
 from time import time
 from contextlib import contextmanager
 
 from webob import Request
+import six
 
 # Actually import time, but close enough to
 # wsgi process start time to use as such
@@ -24,14 +28,14 @@ def uptime():
     return (td.days * 3600 * 24) + td.seconds
 
 
-class StatsRecord(object):
+class StatsRecord:
     def __init__(self, request):
         self._custom_stats = {}
         self.timers = {}
         self.counts = {}
-        self.url = request.environ['PATH_INFO']
+        self.url = six.ensure_text(request.environ['PATH_INFO'])
         if request.environ['QUERY_STRING']:
-            self.url += '?' + request.environ['QUERY_STRING']
+            self.url += '?' + six.ensure_text(request.environ['QUERY_STRING'])
         self.request = request
         # Avoid double-timing things
         self._now_timing = set()
@@ -47,9 +51,10 @@ class StatsRecord(object):
         stats = dict(
             url=self.url,
             uptime=uptime(),
-            timings=dict((k, int(v * 1000))
-                         for k, v in self.timers.iteritems()),
+            timings={k: int(v * 1000)
+                         for k, v in self.timers.items()},
             call_counts=dict(self.counts),
+            pid=os.getpid(),
         )
         stats.update(self._custom_stats)
         return self.to_string(stats)
@@ -75,32 +80,48 @@ class StatsRecord(object):
             yield False
 
 
-class Timer(object):
+class Timer:
     '''Decorator to time a method call'''
-    def __init__(self, timer_name, target_class, *names, **kw):
+    def __init__(self, timer_name: str, target_class: typing.Type | types.ModuleType, *names: str, debug_each_call=True):
+        """
+        :param timer_name: a simple name to record with
+        :param target_class: class or module reference
+        :param names: functions within class/module to time
+        :param debug_each_call: if 'timermiddleware' DEBUG log level, log every time these functions are called
+        """
         self.timer_name = timer_name
         self.target_class = target_class
         self.names = self.get_names(names)
-        self.debug_each_call = kw.get('debug_each_call', True)
+        self.debug_each_call = debug_each_call
 
     def get_names(self, names):
         if '*' not in names:
             return names
         import inspect
+        # TODO: collect @classmethod functions too?
         return [meth[0] for meth in inspect.getmembers(self.target_class,
-            predicate=inspect.ismethod)]
+            predicate=inspect.isfunction)]
 
     def decorate(self, middleware):
         for name in self.names:
+            func_or_method = getattr(self.target_class, name)
+            if isinstance(func_or_method, TimingDecorator):
+                # don't double wrap if code gets run twice
+                continue
+            if not inspect.isfunction(func_or_method) and not inspect.ismethoddescriptor(func_or_method):
+                # TODO: allow @classmethod too?
+                raise TypeError(f"{self.target_class}.{name} is not a function or method "
+                                f"(it's {type(func_or_method)}). For methods, make sure you're targeting a method "
+                                f"from a Class, not from an object instance")
             setattr(self.target_class, name,
-                    TimingDecorator(getattr(self.target_class, name),
+                    TimingDecorator(func_or_method,
                                     self.timer_name,
                                     name,
                                     self.debug_each_call,
                                     middleware))
 
 
-class TimingDecorator(object):
+class TimingDecorator:
     def __init__(self, inner, timer_name, method_name, debug_each_call, middleware):
         self._inner = inner
         self.timer_name = timer_name
@@ -134,7 +155,8 @@ class TimingDecorator(object):
             if instance is not None:
                 call_method += str(type(instance))
             call_method += ' ' + self.method_name
-            call_args = ' args=%s kwargs=%s' % (args, kwargs)
+            call_args = ' args={} kwargs={}'.format(six.ensure_text(repr(args)),
+                                                six.ensure_text(repr(kwargs)))
             if len(call_args) > self.debug_line_length:
                 call_args = call_args[:self.debug_line_length] + '...'
 
@@ -151,7 +173,7 @@ class TimingDecorator(object):
             return retval
 
 
-class GeneratorTimingProxy(object):
+class GeneratorTimingProxy:
     def __init__(self, generator, timer_name, stats, debug_each_call):
         self.generator = generator
         self.timer_name = timer_name
@@ -165,13 +187,14 @@ class GeneratorTimingProxy(object):
         with self.stats.timing(self.timer_name+'.next') as timed_this_one:
             if self.debug_each_call and timed_this_one:
                 log.debug('%s.next call', self.timer_name)
-            return self.generator.next()
+            return next(self.generator)
+    __next__ = next
 
     def __getattr__(self, name):
         return getattr(self.generator, name)
 
 
-class TimerMiddleware(object):
+class TimerMiddleware:
     LISTENERS = []
     stat_record_class = StatsRecord
 
